# -*- coding:utf-8 -*-

# Q's Account Book
# Copyright (c) 2002-2007 HORIKAWA Hisashi. All rights reserved.
#   mailto:vzw00011@nifty.ne.jp
#   http://www.nslabs.jp/

# 勘定科目マスタ

class AcTitle
  attr_reader(
    :code,           # コード 1000〜9999
    :name,           # 名前
    :name_short,     # 携帯電話用
    :ac_type,        # 種類   1 = 資産，2 = 負債，3 = 資本, 4 = 売上高，5 = 営業費用,
                     #        6 = その他の収益, 7 = その他の費用
    :cf_type,        # C/F   1 = 営業，2 = 投資, 3 = 財務, 4 = 現金・現金同等物
    :desc,           # 説明
    :disable         # 新規登録不可でtrue
  )
  
  def initialize(code, name, name_short, ac_type, cf_type, desc)
    raise TypeError if !code.is_a?(Integer) || !name.is_a?(String) ||
                       !ac_type.is_a?(Integer) ||
                       !cf_type.is_a?(Integer) || !desc.is_a?(String)
    raise ArgumentError if code <= 0
    raise ArgumentError, "ac_type = #{ac_type}" if ac_type < 1 || ac_type > 8
    raise ArgumentError, "cf_type = #{cf_type}" if cf_type < 1 || cf_type > 4

    @code = code
    @name = name
    @name_short = name_short
    @ac_type = ac_type == 8 ? 3 : ac_type # backward compatibility 
    @cf_type = cf_type
    @desc = desc
  end

  def AcTitle.parse(str)
    code, name, name_short, ac_type, desc, cf_type = str.split("\t")
    return AcTitle.new(
      code.to_i, e2u(CGI.unescape(name)), e2u(CGI.unescape(name_short)),
      ac_type.to_i, (cf_type || 1).to_i,
      e2u(CGI.unescape(desc || "")))
  end

  def to_s()
    [@code, CGI.escape(u2e(@name)), 
     CGI.escape(u2e((@name_short || "").split(//u)[0..1].join)),
     @ac_type, CGI.escape(u2e(@desc)), @cf_type].join("\t")
  end

  def bs?
    @ac_type <= 3
  end

  def pl?
    @ac_type >= 4
  end
  
  def sign
    return [nil, 1, -1, -1, -1, 1, -1, 1][@ac_type]
  end

  def AcTitle.actype_names
    [nil, '資産', '負債', '純資産', '営業収益', '営業費用',
     'その他の収益', 'その他の費用']
  end

  def AcTitle.cftype_names
    [nil, '営業活動', '投資活動', '財務活動', '現金同等物']
  end

  def AcTitle.group_desc(group)
    case group
    when 1
      "資産は，現金・預金などの財産や，未収入金など将来の入金となるものです。現金や通常使う預金はC/F区分を現金同等物、定期預金などは投資活動にします。"
    when 2
      "負債は，ローン，クレジットカードの残高など，将来何らかの支出（返済）が必要なものです。"
    when 3
      "純資産は，資産から負債を差し引いた，純粋な自身の財産です。「未処分利益」勘定が必要です。"
    when 4
      "営業収益は，通常の給料、家賃収入など，経常的な収入です。ボーナスなど経常的でないものは含みません。"
    when 5
      "営業費用は，経常的に発生する支出です。ときどき（おおむね3ヶ月に1度未満）しか発生しないものは含みません。"
    when 6
      "その他の収益は，金融収益（利息など），臨時的な収入です。"
    when 7
      "その他の費用は，金融費用，税金，臨時的な支出です。"
    end
  end
end

def actitle_find_by_code(ary, code)
  raise TypeError, "code must a Integer" if !code.is_a?(Integer)
  ary.each {|e|
    return e if e.code == code
  }
  return nil
end

def actitle_index_by_code(ary, code)
  ary.each_with_index {|e, idx|
    return idx if e.code == code
  }
  return nil
end

class AcTitleArray < Array
  def code_list
    r = Array.new
    each {|e| r << e.code}
    return r
  end

  def find_by_code(code)
    return actitle_find_by_code(self, code)
  end

  def find_index_by_code(code)
    each_index {|i|
      if self[i].code == code
        return i
      end
    }
    return nil
  end
end

class AcTitleStore
  def AcTitleStore.get(user, db)
    r = AcTitleArray.new
    File.open(db.file_path.ac_titles(user)) {|fp|
      fp.flock(File::LOCK_SH)
      fp.each_line {|line|
        r << AcTitle.parse(line.gsub(/[\r\n]/, ''))
      }
    }
    return r
  end

  def AcTitleStore.store(user, ary)
    ary.each {|e|
      raise TypeError if !e.instance_of?(AcTitle)
    }

    File.open(FilePath.ac_titles(user), File::WRONLY | File::CREAT) {|fp|
      fp.flock(File::LOCK_EX)
      fp.truncate(0)
      ary.each {|e|
        fp.print e.to_s, "\n"
      }
    }
  end
end


class AcChooser
  def initialize(user, titles = nil)
    if titles
      @ac_titles = titles
    else
      @ac_titles = AcTitleStore.get(user)
    end
  end
  
  def get_sel_acitems(name, default, &f)
    s = "<select name=\"#{name}\">\n"
    s << get_sel_acitems_sub(@ac_titles, default, &f)
    s << "</select>"
    return Amrita::SanitizedString[s]
  end

  def get_sel_acitems_with_zero(name, default)
    s = "<select name=\"#{name}\">\n"
    s << get_sel_acitems_sub(@ac_titles, default, true)
    s << "</select>"
    return Amrita::SanitizedString[s]
  end

  # 購買伝票用の勘定科目一覧 
  def get_sel_purchase(name, default)
    ary = []
    @ac_titles.each {|item|
      next if item.ac_type == 2 || item.ac_type == 3 || item.ac_type == 4 ||
              item.ac_type == 6
      next if item.cf_type == 4
      ary << item
    }
    ret = "<select name=\"#{name}\">\n" +
          get_sel_acitems_sub(ary, default) +
          "</select>"
    return Amrita::SanitizedString[ret]
  end

  def get_sel_purchase_with_zero(name, default)
    ary = []
    @ac_titles.each {|item|
      next if item.ac_type == 2 || item.ac_type == 3 || item.ac_type == 4 ||
              item.ac_type == 6
      next if item.cf_type == 4
      ary << item
    }
    ret = "<select name=\"#{name}\">\n" +
          get_sel_acitems_sub(ary, default, true) +
          "</select>"
    return Amrita::SanitizedString[ret]
  end

  private
  def get_sel_acitems_sub(item_ary, default, with_zero = false)
    s = ""
    s << "  <option value=\"0\">　　〃</option>\n" if with_zero
    AcTitle.actype_names.each_with_index {|group, i|
      next if !group
      gr_items = []
      item_ary.each {|item|
        if item.ac_type == i
          if !block_given? || yield(item)
            gr_items << item
          end
        end
      }
      if gr_items.size > 0
        s << "  <optgroup label=\"#{group}\">\n"
        gr_items.each {|item|
          selected = (default == item.code) ? " selected" : ""
          s << "    <option#{selected} value=\"#{item.code}\">" +
               CGI.escapeHTML(item.name) + "</option>\n"
                  # IEはoptgroupと混ぜるときにoptionの終了タグを省略してはいけない
        }
        s << "  </optgroup>\n"
      end
    }
    return s
  end
end
