# -*- coding:utf-8 -*-

# Q's Account Book
# Copyright (c) 2002-2007 HORIKAWA Hisashi. All rights reserved.
#   mailto:vzw00011@nifty.ne.jp
#   http://www.nslabs.jp/

# 仕訳（１行の）

require LIBRARY_DIR + '/datesup.rb'

class JournalRecord
  attr_accessor(
    :seq,       # SEQ                  int   primary key
    :date,      # 日付                 Dateインスタンス
    :num,       # 伝票番号             int
    :dr_code,   # 借方科目コード        int
    :cr_code,   # 貸方科目コード        int
    :amount,    # 金額                 int
    :remarks,   # 相手                 string
    :comment,   # 内容                 string
    :del,       # 削除フラグ trueで削除 true/false 
    :add_time,  # 追加時刻              time
    :del_time   # 削除時刻              time
    )

  def JournalRecord.parse(str)
    date, seq, num, dr_code, cr_code, amount, remarks,
    del, add_time, del_time, comment = str.split("\t")

    obj = JournalRecord.new
    obj.date = Date.parse(date)
    obj.seq = seq.to_i
    obj.num = num.to_i
    obj.dr_code = dr_code.to_i
    obj.cr_code = cr_code.to_i
    obj.amount = amount.to_i
    obj.remarks = e2u(CGI.unescape(remarks))
    obj.comment = comment ? e2u(CGI.unescape(comment)) : ""
    obj.del = del != "" ? true : false
    obj.add_time = add_time && add_time != "" ? ParseDate.get_time(add_time) : nil
    obj.del_time = del_time && del_time != "" ? ParseDate.get_time(del_time) : nil
    return obj
  end

  def validate()
    raise TypeError if !@date.is_a?(Date)
    raise TypeError if !@seq.is_a?(Integer)
    raise TypeError if !@num.is_a?(Integer)
    raise TypeError if !@dr_code.is_a?(Integer)
    raise TypeError if !@cr_code.is_a?(Integer)
    raise TypeError if !@amount.is_a?(Integer)
    raise if @dr_code == 0
    raise if @cr_code == 0
  end

  def to_s()
    validate()

    [@date.to_s, @seq, @num, @dr_code, @cr_code, @amount,
     CGI.escape(u2e(@remarks)),
     (@del ? "1" : ""),
     (@add_time ? CGI.rfc1123_date(@add_time) : ""),
     (@del_time ? CGI.rfc1123_date(@del_time) : ""),
     CGI.escape(u2e(@comment))
    ].join("\t")
  end
end

class JournalStore < StrStore
  undef_method :[]=
  undef_method :clear

  def initialize(userid, db)
    super(db.file_path.journal(userid))
    @db = db
    @userid = userid
  end

  def append(ary)
    raise TypeError if !ary.is_a?(Array)
    in_transaction_wr()

    period = nil
    @db.transaction_ro {
      period = @db.get_addable_period(@userid)
    }

    now = Time.now
    ok = []
    error = []
    ary.each {|r|
      r.add_time = now if !r.add_time
      if r.date < period[0] || r.dr_code == r.cr_code
        error << r
      else
        r.seq = incr_counter(FilePath.seq_counter(@userid)) if !r.seq
        @table[r.seq] = r
        ok << r
      end
    }

    return [ok, error]
  end

  def delete(ary)
    raise TypeError if !ary.is_a?(Array)
    in_transaction_wr()
    
    fixed = nil
    @db.transaction_ro {|db|
      fixed = db.get_profile(@userid).fixed # 締め日
    }

    now = Time.now
    ok = []
    error = []
    del_db = []
    ary.each {|seq|
      raise TypeError if !seq.is_a?(Integer)

      if @table[seq]
        r = @table[seq]
        if r.date <= fixed
          error << r
        else
          r.del_time = now
          ok << r
          del_db << r
          @table.delete(seq)
        end
      end
    }

    File.open(@path + ".deleted", "a") {|fp|
      fp.flock(File::LOCK_EX)
      del_db.each {|r|
        fp.print r.to_s, "\n"
      }
    }

    return [ok, error]
  end

  def update(day, entry_no, ary)
    raise TypeError if !day.is_a?(Date)
    raise TypeError if !ary.is_a?(Array)
    in_transaction_wr()

    period = nil
    @db.transaction_ro { period = @db.get_addable_period(@userid) }

    raise if day < period[0] || day > period[1]
    
    now = Time.now
    @table.keys.each {|k|
      r = @table[k]
      @table.delete(r.seq) if r.date == day && r.num == entry_no
    }
    ary.each {|r|
      raise if r.dr_code == r.cr_code
      raise if r.amount == 0
      r.date = day
      r.num = entry_no
      r.seq = incr_counter(FilePath.seq_counter(@userid))
      r.add_time = now
      @table[r.seq] = r
    }
  end

  def each()
    in_transaction()
    @table.each {|k, v|
      yield v
    }
  end

  def find(seq)
    in_transaction
    return @table[seq]
  end

  def find_by_entry_no(date, entry_no)
    in_transaction
    ret = []
    @table.each {|k, v|
      ret << v if v.date == date && v.num == entry_no
    }
    return ret
  end

  private
  def dump_content(io)
    @table.each {|k, v|
      io.print v.to_s, "\n"
    }
  end

  def load_content(io)
    while line = io.gets
      r = JournalRecord.parse(line.gsub(/[\r\n]/, ''))
      @table[r.seq] = r
    end
  end
end
