# -*- coding:utf-8 -*-

# Web Application Framework 'U'
# Copyright (c) 2002-2005 HORIKAWA Hisashi. All rights reserved.
#   mailto:vzw00011@nifty.ne.jp
#   http://www.nslabs.jp/

# このライブラリを使っているプロジェクト：
#    Q's Account Book
#    Q's WebDiary2
#    common/wa-common.rb
#    Q's FotoAlbum

class File
  def File.rel2abs(rel_path, this_file)
    File.expand_path(rel_path, File.dirname(File.expand_path(this_file)))
  end
end

require File.rel2abs("./cgisup.rb", __FILE__)
require File.rel2abs("./strstore.rb", __FILE__)
require File.rel2abs("./csconv.rb", __FILE__)
require File.rel2abs("./injector.rb", __FILE__)
require "digest/md5"


module Wafu  # Web Application Framework 'U'
  class Action
    def initialize(app)
      @app = app
    end

    # subclass must override this method.
    def execute(af)
      raise "no #{self.class}'s execute() method."
    end
  end

  class ActionForm
    def initialize(app)
      @app = app
      @errors = []
      setup()
    end
    attr_reader :errors

    def validate()
      return set_error(nil)
    end

    def setup; end

    def error?
      return errors && errors.size > 0
    end

    def set_error(error_ary)
      if error_ary
        raise TypeError if !error_ary.is_a?(Array)
        @errors += error_ary
      end
      return self
    end
  end

  class Response
    # RFC 2616 HTTP/1.1
    HTTP_STATUS = { 
      100 => 'Continue',
      101 => 'Switching Protocols',
      200 => 'OK',
      201 => 'Created',
      202 => 'Accepted',
      203 => 'Non-Authoritative Information',
      204 => 'No Content',
      205 => 'Reset Content',
      206 => 'Partial Content',
      300 => 'Multiple Choices',
      301 => 'Moved Permanently',
      302 => 'Found',
      303 => 'See Other',
      304 => 'Not Modified',
      305 => 'Use Proxy',
      307 => 'Temporary Redirect',
      400 => 'Bad Request',
      401 => 'Unauthorized',
      402 => 'Payment Required',
      403 => 'Forbidden',
      404 => 'Not Found',
      405 => 'Method Not Allowed',
      406 => 'Not Acceptable',
      407 => 'Proxy Authentication Required',
      408 => 'Request Timeout',
      409 => 'Conflict',
      410 => 'Gone',
      411 => 'Length Required',
      412 => 'Precondition Failed',
      413 => 'Request Entity Too Large',
      414 => 'Request-URI Too Long',
      415 => 'Unsupported Media Type',
      416 => 'Requested Range Not Satisfiable',
      417 => 'Expectation Failed',
      500 => 'Internal Server Error',
      501 => 'Not Implemented',
      502 => 'Bad Gateway',
      503 => 'Service Unavailable',
      504 => 'Gateway Timeout',
      505 => 'HTTP Version Not Supported'
    }

    def initialize(hash = {})
      @header = hash
    end
    attr_reader :header
    attr_accessor :body
    attr_accessor :status_code
  
    def get_http_header(cgi)
      if status_code
        if HTTP_STATUS[status_code]
          header["status"] = "#{status_code} #{HTTP_STATUS[status_code]}"
        else
          raise "unknown status code"
        end
      end
      return cgi.header(header)
    end

    def redirect?
      case status_code
      when 301, 302, 303, 307
        return true
      else
        return false
      end
    end
  end

  LOGON_SUCCESS = 0
  LOGON_RETRY = 1
  LOGON_EXPIRE = 2

  class WebApp
    attr_reader :cgi, :session, :user
    attr_reader :response

    # opt_hash
    #   "session_type"
    #          "none"             no session
    #          "auto"             create session automaticly
    #          "auth_required"    require user auth
    #   "session_option"  CGI::Session's option
    # セッション種別が"auth_required"のときは、Apacheで認証するか、自分で認証するか。
    # 自分で認証するときは、次の値も必要。
    #   "pwdfile"         password file name.
    #   "passwd_store"    a object which has check_passwd().
    def initialize(opt_hash)
      @cgi = CGI.new("html4")

      if RUBY_VERSION == '1.8.1'
        class << @cgi
          undef_method :fieldset
        end
      end

      @wa_options = opt_hash
      @response = Wafu::Response.new(opt_hash["response_option"] || {})
    end

    # userid, pwdが正しいか検査する。
    # 戻り値
    #    ユーザーが存在し、かつパスワードが正しい場合 => LOGON_SUCCESS
    #    ユーザーIDまたはパスワードが違う場合 => LOGON_RETRY
    def check_passwd(userid_, pwd_)
      return LOGON_RETRY if !userid_ || userid_ == "" || !pwd_ || pwd_ == ""

      if @wa_options['passwd_store']
        store = @wa_options['passwd_store']
      elsif defined?(PasswdStore)
        store = PasswdStore.new(@wa_options['pwdfile'])
      else
        raise "must set passwd_store or pwdfile option."
      end

      if store.check_passwd(userid_, pwd_)
        return LOGON_SUCCESS
      else
        return LOGON_RETRY
      end
    end

  # @return [cgi, session, userid]
  def get_session()
    session_option = @wa_options["session_option"]

    is_session_exist = CGI::Session.exist?(cgi, session_option)
    CGI::Session.sweep(session_option)

    session = nil
    r_state = nil
    if @wa_options["session_type"] == "auth_required"
      session = CGI::Session.get(cgi, session_option)
      if !session || !session["user"] || session["user"] == ""
        if ENV["REMOTE_USER"]
          session = CGI::Session.new(cgi, session_option) if !session
          session["user"] = ENV["REMOTE_USER"].dup
        else
          userid_ = cgi.get("user", nil)
          pwd_ = cgi.get("pwd", nil)
          if userid_ && pwd_
            # ログインしたとき
            r_state = check_passwd(userid_, pwd_)
            if r_state == LOGON_SUCCESS
              session = CGI::Session.new(cgi, session_option) if !session
              session["user"] = userid_ if !session["user"]
            else
              return [nil, nil, r_state]
            end
          else
            if is_session_exist
              r_state = LOGON_EXPIRE
            else
              r_state = nil
            end
            return [nil, nil, r_state]
          end
        end
      end
    else
      # session_type == "auto"
      session = CGI::Session.new(cgi, session_option)
    end

    session.update_access_time() 
    return [session, session["user"].untaint, r_state]
  end

  def setup; end
  def teardown; end

  def main()
    raise "no #{self.class}'s main() method."
  end

  def run()
    session_type = @wa_options["session_type"]
    if session_type == "auto" || session_type == "auth_required"
      @session, @user, state = get_session()
      if session_type == "auth_required" && (!@user || @user == "")
        login_form(state)
        exit
      end
    end

    setup()
    main()
    teardown()

    if response.redirect?
      print response.get_http_header(cgi)
      print <<EOF
<html>
<body>
<p>See <a href="#{response.header["Location"]}">#{response.header["Location"]}</a>
</body>
</html>
EOF
    else
      if response && response.body && response.body != ""
        print response.get_http_header(cgi)
        print response.body
      end
    end
  rescue StandardError, ScriptError
    send_error_mail($!.dup, $@.dup)
    internal_error_message_out()
  end

  def dispatch(actions)
    action = nil
    actions.each {|k, pair|
      raise TypeError if !pair.is_a?(Array)
      if cgi.get("action." + k, nil)   # input[type=submit]で、name=action.foo
        action = pair
        break
      elsif cgi.get("action") == k     # hiddenコントロールで、name="action", value="foo"
        action = pair
        break
      end
    }
    action = actions['*'] if !action
    raise ArgumentError if !action

    af = action[0].new(self) if action[0]
    if af
      af.validate()
      if !af.error?
        aklass = action[1]
      else
        aklass = action[2]
      end
    else
      aklass = action[1]
    end
    raise ArgumentError if !aklass.is_a?(Class)

    while aklass.is_a?(Class) && aklass <= Wafu::Action
      aklass = aklass.new(self).execute(af)
    end
  end

  def set_redirect(url)
    raise TypeError if !url.is_a?(String)

    response.status_code = cgi.server_protocol == "HTTP/1.1" ? 303 : 302
    response.header["Location"] = url
  end

  def create_viewstate_id()
    Digest::MD5.new.update(Time.now.to_f.to_s + rand().to_s).hexdigest
  end

  def get_viewstate_data(idstr)
    raise TypeError if idstr && !idstr.is_a?(String)

    return nil if !idstr || idstr == ""
    if session["vs." + idstr] && session["vs." + idstr] != ""
      r = Marshal.load(session["vs." + idstr])
      return r
    else
      return nil
    end
  end

  def set_viewstate_data(data)
    idstr = create_viewstate_id()
    if data
      session["vs." + idstr] = Marshal.dump(data)
    else
      session["vs." + idstr] = nil
    end
    return idstr
  end

  # 取り出したら削除する。viewstateと違う。
  def get_flash(key)
    r = session[key]
    session[key] = nil
    r = nil if r == ""
    return r
  end

  def set_flash(key, value)
    session[key] = value
  end

  def internal_error_message_out()
    print <<EOF
Content-Type:text/html

<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
            "http://www.w3.org/TR/html4/loose.dtd">
<html>
<body>
An exception raised.
<p>#{CGI.escapeHTML($!.inspect)}
<p>
EOF
    $@.each {|x| print CGI.escapeHTML(x), "<br>\n"}
    print <<EOF
<hr>
#{CGI.escapeHTML(cgi.params.inspect)}
</body>
</html>
EOF
  end

  private
  def send_error_mail(err, at)
    raise RuntimeError if !AppConfig[:admin_mail]

    if AppConfig[:prog_encoding] == 'u'
      site_name = CSConv.u2e(AppConfig[:site_name])
    else
      site_name = AppConfig[:site_name]
    end

    header = {"From" => AppConfig[:prog_mail],
              "To" => AppConfig[:admin_mail],
              "Subject" => "#{site_name}: エラー発生",
              "MIME-Version" => "1.0",
              "Content-Type" => "text/plain; charset=ISO-2022-JP"}
    mailbody = <<EOF
#{site_name}: エラー発生

<date>
#{Time.now}
</date>

<userid>
#{@user}
</userid>

<ua>
#{ENV["HTTP_USER_AGENT"]}
</ua>

<message>
#{err.inspect}
</message>

<backtrace>
EOF
    at.each {|x| mailbody += x + "\n"}
    mailbody += "</backtrace>\n"

    send_mail(header, mailbody)
  end
end

end # module Wafu

